{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "eb0cd743",
   "metadata": {},
   "source": [
    "# Hello World in PyTorch"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "85cbb004",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Imports torch\n",
    "import torch\n",
    "from torch import nn"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "477ef499",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines our first network\n",
    "class FirstNetwork(nn.Module):\n",
    "    def __init__(self, multiplier):\n",
    "        super().__init__()\n",
    "        self.multiplier = nn.Parameter(torch.tensor(float(multiplier)))\n",
    "    \n",
    "    def forward(self, x):\n",
    "        return self.multiplier * x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "41568a63",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Initializes FirstNetwork\n",
    "network = FirstNetwork(2)\n",
    "print(network)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dee9bef5",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Runs the network on an input and gets the result\n",
    "network(4).item()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3434a1cc",
   "metadata": {},
   "source": [
    "# Logistic Regression: Defining the Model"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dfecc491",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines a logistic regression model\n",
    "class LogisticRegression(nn.Module):\n",
    "    def __init__(self, input_features, class_num):\n",
    "        super().__init__()\n",
    "        self.weights = nn.Linear(input_features, class_num)\n",
    "    \n",
    "    def forward(self, x):\n",
    "        return self.weights(x)\n",
    "    \n",
    "    def get_probabilities(self, x):\n",
    "        p = nn.functional.softmax(self(x), dim=-1)\n",
    "        return p.detach().numpy()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c924a76d",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Initializes LogisticRegression with 3 features and 2 classes\n",
    "lr_network = LogisticRegression(3, 2)\n",
    "print(lr_network)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e9eaf694",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines some fake data for testing\n",
    "# 2 examples with 3 features each\n",
    "test_data = torch.tensor([[1.0, 2.0, 3.0],\n",
    "                         [3.0, 2.0, 1.0]])\n",
    "print(test_data)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6e1a0712",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Gets the probabilities for each example for each class\n",
    "lr_network.get_probabilities(test_data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "594e91a7",
   "metadata": {},
   "source": [
    "# Logistic Regression: Defining the Loss Function"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "561d2a4e",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines the loss function\n",
    "loss_func = nn.CrossEntropyLoss()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a4c5a695",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines some fake labels for the fake data\n",
    "test_labels = torch.tensor([1, 1])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d2fafa15",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Calls the network on the test data (using forward)\n",
    "output = lr_network(test_data)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ba808f04",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Checks the shape of the output and labels\n",
    "# (Just for debugging)\n",
    "print(output.shape)\n",
    "print(test_labels.shape)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1bcb2afe",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Computes the average loss over examples\n",
    "loss_func(output, test_labels).item()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "51ee1887",
   "metadata": {},
   "source": [
    "# Logistic Regression: Defining the Data"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "eae1bc4b",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines some random train and test data, where the positive class has\n",
    "# higher average value for the first feature\n",
    "# \n",
    "# Don't worry too much about this code\n",
    "import numpy as np\n",
    "from numpy.random import default_rng\n",
    "X_train = default_rng(0).standard_normal([100, 3])\n",
    "X_test = default_rng(1).standard_normal([100, 3])\n",
    "for i in range(50):\n",
    "    X_train[i, 0] += 0.5\n",
    "    X_test[i, 0] += 0.5\n",
    "Y_train = np.concatenate([np.ones(50), np.zeros(50)])\n",
    "Y_test = np.concatenate([np.ones(50), np.zeros(50)])\n",
    "\n",
    "print(X_train)\n",
    "print()\n",
    "print(Y_train)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f67f923b",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Wraps the data in PyTorch datasets\n",
    "from torch.utils.data import TensorDataset\n",
    "dataset_train = TensorDataset(torch.tensor(X_train, dtype=torch.float32),\n",
    "                              torch.tensor(Y_train, dtype=torch.long))\n",
    "\n",
    "dataset_test = TensorDataset(torch.tensor(X_test, dtype=torch.float32),\n",
    "                             torch.tensor(Y_test, dtype=torch.long))\n",
    "\n",
    "print(len(dataset_train))\n",
    "print(len(dataset_test))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "327567c4",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Creates data loaders to batch up the data\n",
    "from torch.utils.data import DataLoader\n",
    "dataloader_train = DataLoader(dataset_train, batch_size=10, shuffle=True)\n",
    "dataloader_test = DataLoader(dataset_test, batch_size=10, shuffle=False)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e75d6e43",
   "metadata": {
    "scrolled": true
   },
   "outputs": [],
   "source": [
    "for X, Y in dataloader_train:\n",
    "    print(X)\n",
    "    print(Y)\n",
    "    print()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "14096f45",
   "metadata": {},
   "source": [
    "# Logistic Regression: Training"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "986d0152",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Creates an optimizer\n",
    "from torch.optim import SGD\n",
    "lr_optimizer = SGD(lr_network.parameters(), lr=0.0001)\n",
    "print(lr_optimizer)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "15a596ba",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Iteratively minimizes training loss\n",
    "epochs = 500\n",
    "lr_network.train()\n",
    "\n",
    "for _ in range(epochs):\n",
    "    for X, Y in dataloader_train:\n",
    "        # Zeros out gradient information\n",
    "        lr_optimizer.zero_grad()\n",
    "        \n",
    "        # Computes the training loss on the batch\n",
    "        outputs = lr_network(X)\n",
    "        loss = loss_func(outputs, Y)\n",
    "        \n",
    "        # Computes gradient\n",
    "        loss.backward()\n",
    "        \n",
    "        # Takes optimization step\n",
    "        lr_optimizer.step()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0581ab65-4343-460d-861b-be87254d4474",
   "metadata": {},
   "outputs": [],
   "source": [
    "lr_network.weights.weight"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3a8bc3f4-150a-4166-81a9-08f2e59adbb3",
   "metadata": {},
   "outputs": [],
   "source": [
    "lr_network.weights.weight"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1eba5f93-623e-4006-968b-414568cbacb2",
   "metadata": {},
   "outputs": [],
   "source": [
    "lr_network.weights.weight"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "593d961b",
   "metadata": {},
   "source": [
    "# Logistic Regression: Testing"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9359d9f3",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Computes test loss\n",
    "lr_test_loss = 0\n",
    "lr_network.eval()\n",
    "\n",
    "with torch.no_grad():\n",
    "    for X, Y in dataloader_test:\n",
    "        # Computes the training loss on the batch\n",
    "            outputs = lr_network(X)\n",
    "            loss = loss_func(outputs, Y)\n",
    "\n",
    "            lr_test_loss += loss.item() * X.shape[0]\n",
    "\n",
    "print(lr_test_loss)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c795fafe",
   "metadata": {},
   "source": [
    "# From Logistic Regression to Multilayer Perceptron"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5d2ed118",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines a multilayer perceptron model\n",
    "class MLP(nn.Module):\n",
    "    def __init__(self, input_features, hidden_size, class_num):\n",
    "        super().__init__()\n",
    "        self.layer1 = nn.Linear(input_features, hidden_size)\n",
    "        self.activation1 = nn.Sigmoid()\n",
    "        self.layer2 = nn.Linear(hidden_size, class_num)\n",
    "    \n",
    "    def forward(self, x):\n",
    "        out = self.layer1(x)\n",
    "        out = self.activation1(out)\n",
    "        out = self.layer2(out)\n",
    "        return out\n",
    "    \n",
    "    def get_probabilities(self, x):\n",
    "        p = nn.functional.softmax(self(x), dim=-1)\n",
    "        return p.detach().numpy()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "91f534ab",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Initializes MLP with 3 features, 10 hidden units, and 2 classes\n",
    "mlp_network = MLP(3, 10, 2)\n",
    "print(mlp_network)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "55f84e87",
   "metadata": {},
   "source": [
    "# Comparison"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8099dc90",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Trains the MLP\n",
    "mlp_optimizer = SGD(mlp_network.parameters(), lr=0.0001)\n",
    "mlp_network.train()\n",
    "\n",
    "for _ in range(epochs):\n",
    "    for X, Y in dataloader_train:\n",
    "        # Zeros out gradient information\n",
    "        mlp_optimizer.zero_grad()\n",
    "        \n",
    "        # Computes the training loss on the batch\n",
    "        outputs = mlp_network(X)\n",
    "        loss = loss_func(outputs, Y)\n",
    "        \n",
    "        # Computes gradient\n",
    "        loss.backward()\n",
    "        \n",
    "        # Takes optimization step\n",
    "        mlp_optimizer.step()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e8aae846",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Computes test loss for MLP\n",
    "mlp_test_loss = 0\n",
    "mlp_network.eval()\n",
    "\n",
    "with torch.no_grad():\n",
    "    for X, Y in dataloader_test:\n",
    "        # Computes the loss on the batch\n",
    "        outputs = mlp_network(X)\n",
    "        loss = loss_func(outputs, Y)\n",
    "        \n",
    "        mlp_test_loss += loss.item() * X.shape[0]\n",
    "\n",
    "print(mlp_test_loss)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1a95a9c8",
   "metadata": {},
   "outputs": [],
   "source": [
    "if mlp_test_loss < lr_test_loss:\n",
    "    print(\"MLP has lower test loss!\")\n",
    "else:\n",
    "    print(\"LogisticRegression has lower test loss!\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "16114ede",
   "metadata": {},
   "source": [
    "# Defining Convolutional Layers in PyTorch"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ba31ac66",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines an input tensor of 10 x 10 \"images\" with 2 examples 3 color channels\n",
    "color_channels = 3\n",
    "imgs = default_rng(2).standard_normal([2, color_channels, 10, 10])\n",
    "imgs = torch.tensor(imgs, dtype=torch.float32)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a292c434",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Defines some 2d convolutional layers and compares the shapes of the outputs\n",
    "out_channels = 3\n",
    "kernel_size = 5\n",
    "stride = 2\n",
    "\n",
    "conv = nn.Conv2d(\n",
    "    color_channels, out_channels, kernel_size=kernel_size, stride=stride\n",
    ")\n",
    "print(conv)\n",
    "print(conv(imgs).shape)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fe1cc6cd",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Flattens the output so that we can treat features as a vector\n",
    "flatten = nn.Flatten()\n",
    "print(flatten(conv(imgs)).shape)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c5a6d310",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.14"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
